
#ifdef OBSERVEDDATA_WORKING
#error "Recursive inclusion!"
#endif

#ifndef OBSERVEDDATA_H
#define OBSERVEDDATA_H
#define OBSERVEDDATA_WORKING

#include <list>
#include <set>
#include <typeinfo>
#include "tclinterGen.h"
#include "VLObjectPointer.h"


class LanguageElementTypeDescriptor: public Tclinter_Object {
 private:
  std::list <LanguageElementTypeDescriptor*> parents;
  std::list <LanguageElementTypeDescriptor*> children;
  char *nameOfType;
  int isAbstract;
  
 public:
  LanguageElementTypeDescriptor(char *aNameOfType, int aIsAbstract) { 
    nameOfType = aNameOfType;
    isAbstract = aIsAbstract;
  }

  int getIsAbstract() {
    return isAbstract;
  }

  virtual char *getNameOfType() { 
    return nameOfType; 
  }

  VLList getSubtypes() {
    VLList l;
    collectSubtypes(&l);
    return l;
  }

  VLList getSupertypes() {
    VLList l;
    collectSupertypes(&l);
    return l;
  }

  void collectSubtypes(VLList *l) {
    l->append(VLObjectPointer(this));
    std::list<LanguageElementTypeDescriptor*>::iterator p = children.begin();
    while (p != children.end()) {
      (*p)->collectSubtypes(l);
      p++;
    }    
  }

  void collectSupertypes(VLList *l) {
    l->append(VLObjectPointer(this));
    std::list<LanguageElementTypeDescriptor*>::iterator p = parents.begin();
    while (p != parents.end()) {
      (*p)->collectSupertypes(l);
      p++;
    }    
  }

  bool isSubtypeOf(LanguageElementTypeDescriptor *other) {
    if (other==this) return 1;
    std::list<LanguageElementTypeDescriptor*>::iterator p = parents.begin();
    while (p != parents.end()) {
      if ((*p)->isSubtypeOf(other)) return 1;
      p++;
    }
    return 0;    
  }

  void addParent(LanguageElementTypeDescriptor *parent) {
    parents.push_back(parent);
    parent->addChild(this);
  }

  void addChild(LanguageElementTypeDescriptor *child) {
    children.push_back(child);
  }
};



class ChangeHandler: public Tclinter_Object {
 private:
  VLValue cmd;
 public:
  ChangeHandler(VLValue aCmd) { cmd=aCmd; }
  void invoke() { cmd.eval(); }
};


// Das Datenmodell ist aus Objekten zusammengesetzt, die alle der Klasse
// StructureObject angehoeren.

class StructureObject: public Tclinter_Object {
 private:
  std::list<ChangeHandler *> changeHandler;
 protected:
  StructureObject *parent;
 public:
  StructureObject() { parent = NULL; }
  StructureObject(StructureObject *aParent) { parent = aParent; }

  virtual bool isVirtualLanguageObject() { return 0; }
  virtual bool isRelationObject() { return 0; }
  virtual bool isLanguageObject() { return 0; }
  virtual bool isEOLObject() { return 0; }

  virtual VLList getSubElements() { return VLList(); }
  virtual StructureObject *getParent() { return parent; }

  StructureObject *getRoot() {
    StructureObject *result = this;
    while (result->getParent() != NULL) result=result->getParent();
    return result;
  }

  bool isAncestorOf(StructureObject *obj) {
   /* true, falls this vorfahre von obj ist. */

    while (obj != NULL) {
      if (obj==this) return 1;
      obj = obj->getParent();
    }
    return 0;      
  }

  virtual char *getClassName() = 0;
  virtual VLString getInfo() { return VLString("not implemented"); }

  void propagateChange() { 
    std::list<ChangeHandler *>::iterator p = changeHandler.begin();
    while (p != changeHandler.end()) {
      (*p)->invoke();
      p++;
    }
    if (parent) parent->propagateChange();
  }

  void addChangeHandler(ChangeHandler *x) { changeHandler.push_back(x); }
  void removeChangeHandler(ChangeHandler *x) { changeHandler.remove(x); }
};



// Die Klasse LanguageElement ist die Oberklasse aller wirklichen
// Objekte der Sprache (daneben gibt es noch Relationen).

class LanguageObject;

class VirtualLanguageObject: public StructureObject {
 public:
  VirtualLanguageObject(StructureObject *aParent) : StructureObject(aParent) {}
  VirtualLanguageObject() : StructureObject() {}
  virtual bool isVirtualLanguageObject() { return 1; }
  virtual void insertBefore(LanguageObject *elem) = 0;
  virtual VirtualLanguageObject *getSuccessor() = 0;
  virtual VirtualLanguageObject *getPredecessor() = 0;
};


// Die Klasse LanguageElement ist die Oberklasse aller wirklichen
// Objekte der Sprache (daneben gibt es noch Relationen).

class ObservedRef;

class LanguageObject: public VirtualLanguageObject {
private:
  std::set<ObservedRef*> *referer; // Objekte, die eine Referenz auf dieses Objekt besitzen

 public:
  LanguageObject() { referer = new std::set<ObservedRef*>(); }

  virtual bool isLanguageObject() { return 1; }
  virtual LanguageElementTypeDescriptor *getTypeId() = 0;

  bool canExtract();
  void extract();
  VLString getInfo() { return VLString(getClassName()); }
  virtual void insertBefore(LanguageObject *elem);
  VirtualLanguageObject *getSuccessor();
  VirtualLanguageObject *getPredecessor();
  void setParent(StructureObject *aParent) { parent=aParent; }

  void addReferer(ObservedRef *obj) { referer->insert(obj); }
  void removeReferer(ObservedRef *obj) { referer->erase(obj); }
  VLList getRefererList();

};



class EOLObject: public VirtualLanguageObject {
 public:
  EOLObject(StructureObject *aParent) : VirtualLanguageObject(aParent) {}
  virtual bool isEOLObject() { return 1; }
  virtual void insertBefore(LanguageObject *elem);
  char *getClassName() { return "EOLObject"; }
  VirtualLanguageObject *getSuccessor();
  VirtualLanguageObject *getPredecessor();
};


// Objekte der Klasse RelationObject repraesentieren eine Relation
// zwischen Objekte der Klasse LanguageObject.
// Jedes RelationObject hat einen bestimmten Besitzer und
// damit eine bestimmte Rolle.

class RelationObject: public StructureObject {
 private:
  char *role;
 public:
  RelationObject(StructureObject *aParent, char *aRole) : StructureObject(aParent) { role=aRole; }
  virtual bool isRelationObject() { return 1; }
  virtual bool isAggregate() = 0;
  virtual char *getRole() { return role; }
};




class AggregateRelation: public RelationObject {
 private:
  char multiplicityKind;
  std::list<LanguageObject*> *data;
  typedef std::list<LanguageObject*>::iterator iterator;
  LanguageElementTypeDescriptor *acceptedType;
  EOLObject *eolObject;

 public:
  AggregateRelation(StructureObject *aParent, char *role, LanguageElementTypeDescriptor *aAcceptedType, char aMultiplicityKind) : RelationObject(aParent, role) {
    data = new std::list<LanguageObject*>();
    acceptedType = aAcceptedType;
    eolObject = new EOLObject(this);
    multiplicityKind = aMultiplicityKind;
  }

  bool isAggregate() { return 1; }

  bool maxMultiplicityIsOne() { return multiplicityKind=='?' || multiplicityKind=='!' || multiplicityKind=='%'; }

  bool minMultiplicityIsOne() { return multiplicityKind=='%'; }

  bool minProcessorMultiplicityIsOne() { return multiplicityKind=='!' || multiplicityKind=='%'; }

  void checkType(LanguageObject *elem) {
    if (!elem->getTypeId()->isSubtypeOf(acceptedType)) {
      throw VLString("Unexpected type: need ", 
            VLString(acceptedType->getNameOfType(),
            VLString(", got ",
            VLString(elem->getTypeId()->getNameOfType() ))));
    }
  }

  LanguageElementTypeDescriptor *getAcceptedType() {
    return acceptedType;
  }

  iterator find(LanguageObject *reference) {
    iterator p = data->begin();
    while (p != data->end()) {
      if ((*p) == reference) return p;
      p++;
    }
    throw VLString("Can't find element!");
  }


  EOLObject *getEOL() { return eolObject; }

  VirtualLanguageObject *getSuccessor(LanguageObject *elem) {
    iterator p = find(elem);
    p++;
    if (p==data->end()) {
      return eolObject;
    } else {
      return *p;
    } 
  }

  VirtualLanguageObject *getPredecessor(LanguageObject *elem) {
    iterator p = find(elem);
    if (p==data->begin()) {
      return eolObject;
    } else {
      p--;
      return *p;
    } 
  }


  VirtualLanguageObject *getFirst() {
    iterator p = data->begin();
    if (p==data->end()) {
      return eolObject;
    } else {
      return *p;
    }    
  }

  void insertBefore(LanguageObject *elem, LanguageObject *reference) {
    insertBefore(elem, find(reference));
  }

  void insertAtEnd(LanguageObject *elem) {
    insertBefore(elem, data->end());
  }

  void insertBefore(LanguageObject *elem, iterator p) {
    if (elem->getParent()) {
       throw VLString("Can't insert; element has a parent");
    }
    if (elem->isAncestorOf(this)) {
        throw VLString("Can't insert; would induce a circle");
    }
    checkType(elem);
    data->insert(p, elem);
    elem->setParent(this);
    propagateChange();
  }

  void remove(LanguageObject *elem) {
    data->remove(elem);
    elem->setParent(NULL);
    propagateChange();
  }

  VLList getSubElements() {
    VLList result = VLList();
    iterator p = data->begin();
    while (p != data->end()) {
      result.append(VLObjectPointer(*p));
      p++;
    }
    return result;    
  }

  char *getClassName() { return "AggregateRelation"; }

  VLString getInfo() { return VLString(getClassName(), VLString(":", VLString(acceptedType->getNameOfType()))); }

};

typedef AggregateRelation *AggregateRelationPtr;


class ObservedType: public RelationObject {
 private:
  VLValue *valuePtr;
  bool nullAllowed;
 public:

  ObservedType(StructureObject *aParent, char *role, bool aNullAllowed, VLValue *aValuePtr) : RelationObject(aParent, role) {
    valuePtr = aValuePtr;
    nullAllowed = aNullAllowed;
  }

  bool isAggregate() { return 0; }

  bool getNullAllowed() { return nullAllowed; }

  virtual VLValue getValue() {
    return *valuePtr;
  }

  virtual void setValue(VLValue aValue) {
    *valuePtr = aValue;
    propagateChange();
  }

  char *getClassName() { return "ObservedType"; }

  VLList getSubElements() {
    return VLList();
  }

  VLString getTypeName() {
    char const *p = typeid(*valuePtr).name();
    while(*p >= '0' && *p <= '9') p++;
    // FIXME: This is a hack to get the real classname. I fear this only works with gnu c++.
    if(strcmp(p,"VLObjectPointer")==0) {
        return VLString("VLPointer");
    } 
    else {
      return VLString(p);
    }
  }

  VLString getInfo() { return VLString(getClassName(), VLString(":", getTypeName())); }
};

typedef ObservedType *ObservedTypePtr;


class ObservedRef: public ObservedType {
public:
  ObservedRef(StructureObject *aParent, char *role, bool aNullAllowed, VLValue *aValuePtr) : ObservedType(aParent, role, aNullAllowed, aValuePtr) {}

  LanguageObject *castToPtr(VLValue aValue) {
    if (VLVoidPointer(aValue).isNull()) {
       return NULL;
    }
    LanguageObject *result = dynamic_cast<LanguageObject*>(VLObjectPointer(aValue).getValue());
    if (!result) { 
      throw VLString("Object pointer has invalid type (not LanguageObject)"); 
    }
    return result;
  }

  void setValue(VLValue aValue) {    
    setRefValue( castToPtr(aValue) );
  }

  LanguageObject *getRefValue() {
    return castToPtr(getValue());
  }

  void setRefValue(LanguageObject *obj) {
    LanguageObject *oldValue = getRefValue();
    if (oldValue) oldValue->removeReferer(this);
    if (obj) obj->addReferer(this);
    ObservedType::setValue(VLObjectPointer(obj));
  }
};

typedef ObservedRef *ObservedRefPtr;



#undef OBSERVEDDATA_WORKING
#endif

