
# Modul: activeCanvas
#
# Requires: vec, misc.
#
# activeCancas ist ein "high level" Canvas inklusive Scrollbalken. Die Abstraktionsebene im Vergeleich
# zum normalen Canvas ist wesentlich hher. Logischen Objekten ("Gruppen") knnen aus mehreren Primitiven zusammengesetzt
# werden. Gruppen knnen interaktive Eigenschaften zugeordnet werden.
#
# activeCanvas untersttz folgende Konpzepte:
#
#  - Gruppen = Mengen von Canvas-Primitiven
#  - Selektion von Gruppen
#  - Zuordnung von Tags zu Gruppen
#  - Drag-and-Drop (eine Gruppe ist dabei das Drag-Objekt, eine andere das Drop-Objekt)
#  - Rotation von Drop-Selektionen (falls mehrere visuell bereinander liegen)
#  - Aktion durch Doppelklick
#  - Kontextmen ffnen durch Rechtsklick
#  - Drag-and-Drop ber Fenstergrenzen hinweg
#
# Es fehlt noch: Mehrfachselektion!
#
# Gruppen besitzen folgende Eigenschaften:
#
#  - Haben eine beliebige Menge von Attribut/Wert - Tags
#  - wissen, auf welche Objekte sie "gedropped" werden knnen und welche Wirkung das hat
#  - Haben einen der folgenden Darstellungszustnde: Normal, Selected, Highlighted
# 
# Canvas-Primitive besitzen folgende Eigenschaften:
#
#  - clickable
#  - dropable
#  - Darstellungszustand
#
# Das Modul ist unabhngig von der Implementierung der zugrundeliegenden abstrakten Struktur, d.h. auch
# Auerhalb von DEViL anwendbar. Gltigkeit von von Drop-Operationen, Eintrge des Kontextmens, sowie 
# Auswirkungen von Operationen knnen Mittels Callback-Funktionen beschrieben werden.


namespace eval ::activeCanvas {
}


###################################################################################### basic operations


proc activeCanvas::removeOptions {toRemove inputList} {
    set result ""
    foreach {attr val} $inputList {
        if { ![elementOf $attr $toRemove] } {
            lappend result $attr $val
        }
    }
    return $result
}


proc activeCanvas::removeUnchangedOptions {optionList changeOptionList} {
    foreach {opt value} $changeOptionList {
        set change($opt) 1
    }
    set result ""
    foreach {opt value} $optionList {
        if { [info exists change($opt)] } {
            lappend result $opt $value
        }
    }
    return $result
}


proc activeCanvas::createPrimitive {obj type pos params} {
    upvar 0 ::activeCanvas::$obj data
    set result ""
    if { [procExists ::canvasitems::$type] } {
        foreach {iType iPos iParams} [::canvasitems::$type $pos $params] {
            lappend result [eval $data(canvas) create $iType $iPos $iParams]
        }        
    } else {
        lappend result [eval $data(canvas) create $type $pos $params]
    }
    return $result
}


proc activeCanvas::createObject {obj group type pos args} {
    upvar 0 ::activeCanvas::$obj data
    assert { ![empty $pos] }
    handleArgs [list -normal "" -select "" -highlight "" -tags "" -level "0" -rotate 0]
    lappend arg(tags) "level-$arg(level)"
    foreach cObj [createPrimitive $obj $type $pos [concat [removeOptions "-clickable -dropable" $arg(normal)] -tags [list $arg(tags)]]] {
        set data(obj-$cObj,normalAttrs) [removeUnchangedOptions $arg(normal) [concat $arg(select) $arg(highlight) "-clickable xxx -dropable xxx"]]
        set data(obj-$cObj,selectAttrs) $arg(select)
        set data(obj-$cObj,highlightAttrs) $arg(highlight)
        setTagValue $obj $cObj group $group
        setTagValue $obj $cObj rotate $arg(rotate)
        setElemState $obj $cObj normal
    }
    return $cObj
}


proc activeCanvas::createObject_xxx {obj group type pos args} {
    # kann auch benutzerdefinierte primitive ("canvasitems.tcl") verarbeiten
    if { [procExists ::canvasitems::$type] } {
        foreach {itemtype coords params} [::canvasitems::$type $pos $args] {
            createObject_internal $obj $group itemtype $coords $params
        }
    } else {
        createObject_internal $obj $group $type $pos $args
    }        
}



proc activeCanvas::setAttributes {obj cObj list} {
    upvar 0 ::activeCanvas::$obj data

    foreach {attr val} $list {
        switch -- $attr {
            "-clickable" {
                setTagValue $obj $cObj clickable $val
            }
            "-dropable" {
                setTagValue $obj $cObj drop $val
            }
            default {
                $data(canvas) itemconfigure $cObj $attr $val
            }
        }
    }
}

    
proc activeCanvas::setElemState {obj tag state} {
    upvar 0 ::activeCanvas::$obj data
    assert { [elementOf $state "normal select highlight"] }
    foreach cObj [$data(canvas) find withtag $tag] {
        setAttributes $obj $cObj $data(obj-$cObj,normalAttrs)
        setAttributes $obj $cObj $data(obj-$cObj,${state}Attrs)
    }
}


proc activeCanvas::setGroupState {obj group state} {
    upvar 0 ::activeCanvas::$obj data
    upvar 0 ::activeCanvas::$obj-states states
    
    set states($group) $state
    setElemState $obj group-$group $state
}


proc activeCanvas::getGroupsWithState {obj state} {
    upvar 0 ::activeCanvas::$obj data
    upvar 0 ::activeCanvas::$obj-states states

    set result ""
    foreach {group gstate} [array get states] {
        if { [equal $gstate $state] } {
            lappend result $group
        }
    }
    return $result
}


proc activeCanvas::getTagValue {obj cObj tagname} {
    upvar 0 ::activeCanvas::$obj data

    assert { ![empty $cObj] }
    set tags [$data(canvas) itemcget $cObj -tags]
    set index [lsearch $tags "$tagname-*"]
    assert { $index != -1 }
    set tag [lindex $tags $index]
    set value [join [lrange [split $tag "-"] 1 end] "-"]
    #old solution: regexp "^$tagname-(.*)" $tag dummy value
    return $value
}


proc activeCanvas::setTagValue {obj cObj tagname value} {
    upvar 0 ::activeCanvas::$obj data

    set tags [$data(canvas) itemcget $cObj -tags]
    set index [lsearch $tags "$tagname-*"]
    set newtags [lreplace $tags $index $index $tagname-$value]
    $data(canvas) itemconfigure $cObj -tags $newtags
}


proc activeCanvas::hasTag {obj cObj tag} {
    upvar 0 ::activeCanvas::$obj data

    set tags [$data(canvas) itemcget $cObj -tags]
    return [elementOf $tag $tags]
}


proc activeCanvas::getDropTargetGroups {self} {
    # benoetigt, um Aktivierung von der Toolbar-Knoepfe effizienter realisieren
    # zu koennen als mittels "canInsert".

    upvar 0 ::activeCanvas::$self data
    set result ""
    foreach cObj [$data(canvas) find withtag drop-1] {
        lappend result [getTagValue $self $cObj group]
    }
    return $result
}


proc activeCanvas::findObjects {self tag predicate} {
    upvar 0 ::activeCanvas::$self data

    set result ""
    foreach cObj [$data(canvas) find withtag $tag] {
        set group [getTagValue $self $cObj group]
        if { [eval $predicate $group] } {
            lappend result $cObj
        }
    }
    return $result
}


proc activeCanvas::findNearestSet {obj pos tag predicate} {
    upvar 0 ::activeCanvas::$obj data

    set nearestDist "1000000"
    set nearestList ""

    foreach cObj [findObjects $obj $tag $predicate] {
        set group [getTagValue $obj $cObj group]
        set dist [vec::distancePointRect $pos [$data(canvas) bbox $cObj]]
        if { $dist < $nearestDist } {
            set nearestDist $dist
            set nearestList ""
        }
        if { $dist == $nearestDist } {
            lappend nearestList $cObj
        }
    }
    if { ![empty $nearestList] && [getTagValue $obj [lindex $nearestList end] rotate]==0 } {
        return [lindex $nearestList end]
    } else {
        return $nearestList
    }
}


proc activeCanvas::findNearest {obj pos tag predicate} {
    set nearestList [findNearestSet $obj $pos $tag $predicate]
    return [lindex $nearestList end]
}


proc activeCanvas::findObjectAt {obj pos tag} {
    upvar 0 ::activeCanvas::$obj data

    set cObjects [lreverse [eval $data(canvas) find overlapping [vec::makeSquare $pos $cfg::clickBox]]]
    foreach cObj $cObjects {
        if { [hasTag $obj $cObj $tag] } {
            return $cObj
        }
    }
    return ""
}

proc activeCanvas::canvasPos {obj windowPos} {
    # rechnet auf canvas-koordinaten um (scrolling)
    upvar 0 ::activeCanvas::$obj data
    assign "x y" $windowPos
    set result "[$data(canvas) canvasx $x] [$data(canvas) canvasy $y]"
    return $result
}

proc activeCanvas::screenCoords2winCoords {obj absPos} {
    upvar 0 ::activeCanvas::$obj data
    assign "x y" $absPos
    set dx [winfo rootx $data(canvas)]
    set dy [winfo rooty $data(canvas)]
    return "[expr $x - $dx] [expr $y - $dy]" 
}



###################################################################################### abstract interaction

proc activeCanvas::openContextMenu {obj windowPos X Y} {
    upvar 0 ::activeCanvas::$obj data
    #on_buttonPress1 $obj $windowPos
    selectPointedElement $obj $windowPos
    set group [getSelection $obj]
    if { ![empty $group] } {
        eval $data(onContextMenu,$group) $group $X $Y
    }
}

proc activeCanvas::startDoubleClickAction {obj} {
    upvar 0 ::activeCanvas::$obj data

    set group $data(clickedGroup)
    if { ![empty $group] } {
        if { ![equal $data(onDoubleClick,$group) "activeCanvas::onDoubleClick_nothing"] } {
            eval $data(onDoubleClick,$group) $group
            return
        }
    }
    set group [getSelection $obj]
    if { ![empty $group] } {
        eval $data(onDoubleClick,$group) $group
    }
}

proc activeCanvas::getPointedElement {obj windowPos} {
    upvar 0 ::activeCanvas::$obj data
    set pos [canvasPos $obj $windowPos]
    set data(clickpos) $pos
    set data(dropPos) $pos
    set data(drag) 0
    set data(dropgroup) ""

    set data(clickedCObj) [findObjectAt $obj $pos clickable-1]
    if { ![empty $data(clickedCObj)] } {
        set data(clickedGroup) [getTagValue $obj $data(clickedCObj) group]
        set data(acceptDropPredicate) "$data(acceptDropFkt,$data(clickedGroup)) $obj $data(clickedGroup)"
        set data(onDrop) $data(onDrop,$data(clickedGroup))
    } else {
        set data(clickedGroup) ""
    }
    return $data(clickedGroup)
}


proc activeCanvas::selectPointedElement {obj windowPos} {
    setSelection $obj [getPointedElement $obj $windowPos]
}


proc activeCanvas::updateDropPointSet {obj pos} {
    upvar 0 ::activeCanvas::$obj data
    if { ![empty $pos] } {
        set nearestSet [findNearestSet $obj $pos drop-1 $data(acceptDropPredicate)]
    } else {
        set nearestSet ""
    }
    set data(dropPointMousePos) $pos
    if { ![equal $nearestSet $data(selectedDropPointSet)] } {
        set data(selectedDropPointSet) $nearestSet
        set data(dropPointIndex) 0
        highlightDropPoint $obj
        rotateDropPoint $obj
    } else {
        highlightDropPoint $obj
    }
}


proc activeCanvas::rotateDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
    if { ![empty $data(dropPointRotationTimer)] } {
        after cancel $data(dropPointRotationTimer)
    }
    if { ![empty $data(selectedDropPointSet)] } {
        set len [llength $data(selectedDropPointSet)]
        set data(dropPointIndex) [expr ($data(dropPointIndex)+$len+-1) % $len]
        set data(dropPointRotationTimer) [after $::cfg::rotationDelay [list activeCanvas::rotateDropPoint $obj]]
        highlightDropPoint $obj
    }        
}    

proc activeCanvas::unhighlightDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data

    foreach group [getGroupsWithState $obj highlight] {
        setGroupState $obj $group normal
    }
    $data(canvas) delete group-temp
    set data(dropgroup) ""

}

proc activeCanvas::highlightDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
    unhighlightDropPoint $obj
    if { [empty $data(selectedDropPointSet)] } {
        return
    }
    set pos $data(dropPointMousePos)
    set nearest [lindex $data(selectedDropPointSet) $data(dropPointIndex)]

    set group [getTagValue $obj $nearest group]
    setGroupState $obj $group highlight
    set data(dropgroup) $group
    set oldpos $data(dropPos)
    set data(dropPos) $pos
    set data(dropDelta) [vec::sub $pos $data(clickpos)]
    eval $data(onVisualDrop,$group) [list $obj $group $data(clickpos) $pos [vec::sub $pos $oldpos]]
}


proc activeCanvas::selectDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data

    if { ![empty $data(dropgroup)] } {
        eval $data(onDrop) [list $obj $data(clickedGroup) $data(dropgroup) $data(dropPos) $data(dropDelta)]
    }
    updateDropPointSet $obj ""
}


# XXX
proc activeCanvas::startAbstractDrop {obj acceptDropPredicate onDrop clickedGroup} {
    upvar 0 ::activeCanvas::$obj data
    set data(acceptDropPredicate) $acceptDropPredicate
    set data(dropPos) "0 0"
    set data(clickpos) "0 0"
    set data(clickedGroup) $clickedGroup
    set data(onDrop) $onDrop

    # 2. updateDropPointSet
    # 3. selectDropPoint
}



###################################################################################### concrete interaction


# Werte von $data(interactionState)   |  Verfuegbare Variablen
# ---------------------------------------------------------------------------
#   objectSelection
#   maybeDraging                         clickpos, clickedCObj, clickedGroup
#   draging                              clickpos, clickedCObj, clickedGroup, dropPos, dropgroup
#   inspointSelection                    clickpos, clickedCObj, clickedGroup, dropPos, dropgroup


proc activeCanvas::setIaState {obj state} {
    upvar 0 ::activeCanvas::$obj data
    set data(interactionState) $state
}


proc activeCanvas::canInsert {self acceptDropPredicate} {
    upvar 0 ::activeCanvas::$self data
    set found [findObjects $self drop-1 $acceptDropPredicate]
    return [expr [llength $found]>0]
}


proc activeCanvas::startObjectInsertion {obj acceptDropPredicate onDrop} {
    upvar 0 ::activeCanvas::$obj data
    returnToObjectSelection $obj
    setIaState $obj inspointSelection
    set data(acceptDropPredicate) $acceptDropPredicate
    set data(dropPos) "0 0"
    set data(clickpos) "0 0"
    set data(clickedGroup) "NOGROUP"
    set data(onDrop) $onDrop
}


proc activeCanvas::returnToObjectSelection {obj} {
    upvar 0 ::activeCanvas::$obj data
    updateDropPointSet $obj ""
    setIaState $obj objectSelection
}


proc activeCanvas::on_buttonPress3 {obj windowPos X Y} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        objectSelection {
            openContextMenu $obj $windowPos $X $Y
        }
        default {
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_doubleClick {obj windowPos} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        objectSelection {
            startDoubleClickAction $obj
        }
        default {
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_buttonPress1 {obj windowPos} {
    upvar 0 ::activeCanvas::$obj data
    switch -regexp $data(interactionState) {
        objectSelection {
            selectPointedElement $obj $windowPos
            if { ![empty $data(clickedGroup)] } {
                setIaState $obj maybeDraging
                set data(interWindowDragDest) $obj
            }
        }
        inspointSelection {
            selectDropPoint $obj
        }
        default {
            log::error "on_buttonPress1: Invalid state detected: $data(interactionState)"
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_b1Motion {obj windowPos absPos} {
    upvar 0 ::activeCanvas::$obj data

    set pos [canvasPos $obj $windowPos]
    
    switch $data(interactionState) {
        objectSelection {
            return
        }
        inspointSelection {
            updateDropPointSet $obj $pos
        }
        draging {
            set destCanvas [eval winfo containing $absPos]
            if { [info exists activeCanvas::activeCanvasId($destCanvas)] } {
                set destCanvas $activeCanvas::activeCanvasId($destCanvas)
            } else {
                set destCanvas $obj
            }
            if { ![equal $destCanvas $data(interWindowDragDest)] } {
                updateDropPointSet $data(interWindowDragDest) ""
                set data(interWindowDragDest) $destCanvas
                if { ![equal $destCanvas $data(canvas)] } {
                    startAbstractDrop $destCanvas $data(acceptDropPredicate) $data(onDrop) $data(clickedGroup)
                }
            }
            updateDropPointSet $destCanvas [canvasPos $destCanvas [screenCoords2winCoords $destCanvas $absPos]]
        }
        maybeDraging {
            if { [vec::distance $data(clickpos) $pos] > $cfg::moveTreshold } {
                setIaState $obj draging
                updateDropPointSet $obj $pos
            }
        }
        default {
            log::error "activeCanvas::on_b1Motion: Invalid state detected"
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_buttonRelease {obj} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        maybeDraging {
            setIaState $obj objectSelection
        }            
        draging {
            selectDropPoint $data(interWindowDragDest)
            setIaState $obj objectSelection
        }
    }
}


############################################################################### some default implementations


proc activeCanvas::onVisualDrop_insertionCursor {cursorSize obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    set cursor [$data(canvas) bbox group-$data(clickedGroup)]
    if { [empty $cursor] } {
        set cursor "$pos [vec::add $pos $cursorSize]"
    } else {
        set cursor [vec::move $cursor [vec::sub $pos $startPos]]
    }
    set cursor [vec::moveInside $cursor [$data(canvas) bbox group-$group]]

    createObject $obj temp rectangle $cursor  -normal "-width 2"
    set data(dropPos) [lrange $cursor 0 1]
}


proc activeCanvas::onVisualDrop_moveArrow {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    assign "x y dummy dummy" [vec::moveInside "$pos $pos" [$data(canvas) bbox group-$group]]
    createObject $obj temp line "$startPos $x $y" -normal "-arrow last"
}


proc activeCanvas::onVisualDrop_moveArrowFullScreen {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    createObject $obj temp line "$startPos $pos" -normal "-arrow last"
}


proc activeCanvas::onVisualDrop_createAssoc {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    createObject $obj temp line "$startPos $pos" -normal "-dash {4 4}"
}

proc activeCanvas::onVisualDrop_movePreview {obj group startPos pos delta} {
    # attribute "visualDragParams" of the draged group must be defined!

    upvar 0 ::activeCanvas::$obj data
    assign "dx dy" $delta
    foreach cObj [$data(canvas) find withtag group-$data(clickedGroup)] {
        set newCoords ""
        foreach {x y} [$data(canvas) coords $cObj] {
            lappend newCoords [expr $x+$dx] [expr $y+$dy]
        }
        $data(canvas) coords $cObj $newCoords
    }
    foreach {cObj pointIndex} [getAttr $obj $data(clickedGroup) "visualDragParams"] {
        set coords [$data(canvas) coords $cObj]
        set x [lindex $coords [expr 2*$pointIndex]]
        set y [lindex $coords [expr 2*$pointIndex+1]]
        set x [expr $x+$dx]
        set y [expr $y+$dy]
        $data(canvas) coords $cObj [lreplace $coords [expr 2*$pointIndex] [expr 2*$pointIndex+1] $x $y]
    }
}

proc activeCanvas::onVisualDrop_nothing {obj group startPos pos delta} {
}


proc activeCanvas::acceptDropFkt_all {canvas fromGroup toGroup} {
    return 1
}

proc activeCanvas::acceptDropFkt_none {canvas fromGroup toGroup} {
    return 0
}

proc activeCanvas::onDrop_nothing {canvas fromGroup toGroup absPos delta} {
}

proc activeCanvas::onContextMenu_nothing {group X Y} {
}

proc activeCanvas::onDoubleClick_nothing {group} {
}


########################################################################################## interface


# PUBLIC
proc activeCanvas::create {obj} {
    # Creates a virtual widget with path $obj

    upvar 0 ::activeCanvas::$obj data  

    frame $obj
    canvas $obj.workspace -background white
    CDlg::mkScrollableXY $obj workspace

    set data(canvas) $obj.workspace
    set data(selgroup) ""
    startRedraw $obj

    bind $data(canvas) <ButtonPress-1> [list activeCanvas::on_buttonPress1 $obj "%x %y"]
    bind $data(canvas) <Double-ButtonPress-1> [list activeCanvas::on_doubleClick $obj "%x %y"]
    bind $data(canvas) <ButtonPress-3> [list activeCanvas::on_buttonPress3 $obj "%x %y" %X %Y]
    bind $data(canvas) <Motion> [list activeCanvas::on_b1Motion $obj "%x %y" "%X %Y"]
    bind $data(canvas) <ButtonRelease-1> [list activeCanvas::on_buttonRelease $obj]

    set ::activeCanvas::activeCanvasId($data(canvas)) $obj
}


proc activeCanvas::newElement {obj group type pos args} {
    # Creates a new canvas object of type $type at $pos and associates it to $group.
    # Args: -normal, -select and -highlight define the graphical properties
    # in the three states. Initially an element is in normal state. When it is selected it is in selected state.
    # When it is chosen as drop-target it is highlight state.
    # Beside the properties of the canvas item there are two special properties:
    # -clickable: Only clickable objects can be selected or draged by the mouse pointer.
    # -dropable: Only dropable objects can accept drops (just an optimization)

    handleArgs [list -normal "" -select "" -highlight "" -level 0 -rotate 0 -print 0]

    upvar 0 ::activeCanvas::$obj data
    if { ![info exists data(onDrop,$group)] } {
        # default group definition, in case it is not provided by the user
        defGroup $obj $group
    }
    set res [createObject $obj $group $type $pos -normal $arg(normal) -select $arg(select) -highlight $arg(highlight) -level $arg(level) -rotate $arg(rotate) \
      -tags "print-$arg(print)"]
    return $res
}


proc activeCanvas::defGroup {obj group args} {
    # Defines the properties of a group. 
    # -acceptDropFkt: Function that determines which drop-targets are accepted. 
    # -onDrop: Function which is called when something is droped at this group
    # -dominatorGroup: When this group is selected it transfers the the selection to its dominator
    #    When the dominator is selected, all dominated groups are in selected mode.
    # -onVisualDrop: Function, that is executed when this group is highlighted for a drop
    # -onContextMenu: Function, that is executed when the right mouse button is pressed
    # -onDoubleClick: Function, that is executed when the left mouse button is double clicked

    upvar 0 ::activeCanvas::$obj data
    handleArgs [list -acceptDropFkt "activeCanvas::acceptDropFkt_all" -onDrop "activeCanvas::onDrop_nothing" \
      -dominatorGroup "" -onVisualDrop "activeCanvas::onVisualDrop_nothing" \
      -onContextMenu activeCanvas::onContextMenu_nothing -onDoubleClick activeCanvas::onDoubleClick_nothing]

    set data(defined,$group) 1
    set data(onDrop,$group) $arg(onDrop)
    set data(acceptDropFkt,$group) $arg(acceptDropFkt)
    set data(onVisualDrop,$group) $arg(onVisualDrop)
    set data(onContextMenu,$group) $arg(onContextMenu)
    set data(onDoubleClick,$group) $arg(onDoubleClick)
    set data(dominatorGroup,$group) $arg(dominatorGroup)
    if { ![empty $arg(dominatorGroup)] } {
        lappend data(subgroups,$arg(dominatorGroup)) $group
        # former lappend2
    }
}

# PUBLIC 
proc activeCanvas::setAttr {obj group attr value} {
    upvar 0 ::activeCanvas::$obj data
    set data(attr-$attr,$group) $value
}

# PUBLIC 
proc activeCanvas::getAttr {obj group attr args} {
    upvar 0 ::activeCanvas::$obj data
    handleArgs [list -dest ""]
    if { ![empty $arg(dest)] } {
        upvar 1 $arg(dest) dest
        if { [info exists data(attr-$attr,$group)] } {
            set dest $data(attr-$attr,$group)
            return 1
        } else {
            return 0
        }
    } else {            
        return $data(attr-$attr,$group)
    }
}


# PUBLIC
proc activeCanvas::containsGroup {obj group} {
    upvar 0 ::activeCanvas::$obj data
    return [info exists data(defined,$group)]
}

# PUBLIC
proc activeCanvas::trySetSelectionAndScroll {obj group} {
    if { [containsGroup $obj $group] } {
        setSelectionAndScroll $obj $group
    }
}


# PUBLIC
proc activeCanvas::setSelectionAndScroll {obj group} {
    ensureVisible $obj $group
    setSelection $obj $group
}


# PUBLIC
proc activeCanvas::setSelectionWithEffect {obj group} {
    ensureVisible $obj $group
    setSelection $obj ""
    after 50 [list activeCanvas::setSelection $obj $group]
    after 100 [list activeCanvas::setSelection $obj ""]
    after 150 [list activeCanvas::setSelection $obj $group]
}


# PUBLIC
proc activeCanvas::setSelection {obj group} {
    upvar 0 ::activeCanvas::$obj data

    foreach old [getGroupsWithState $obj select] {
        setGroupState $obj $old normal
    }

    if { ![empty $group] } {
        if { ![empty $data(dominatorGroup,$group)] } {
            set group $data(dominatorGroup,$group)
        }
    }
    set data(selgroup) $group

    if { ![empty $data(selgroup)] } {
        setGroupState $obj $data(selgroup) select
        if { [info exists data(subgroups,$data(selgroup))] } {
            foreach x $data(subgroups,$data(selgroup)) {
                setGroupState $obj $x select
            }
        }
    }
}

# PUBLIC
proc activeCanvas::getSelection {obj} {
    upvar 0 ::activeCanvas::$obj data
    return $data(selgroup)
}

# PUBLIC
proc activeCanvas::getBoundingBox {obj elemList} {

    upvar 0 ::activeCanvas::$obj data
    return [eval $data(canvas) bbox $elemList]
}


proc activeCanvas::ensureVisible {obj node} {
    upvar 0 ::activeCanvas::$obj data

    set bbox [eval $data(canvas) bbox group-$node]
    set region [$data(canvas) cget -scrollregion]

    # fr x
    set mid [expr ([lindex $bbox 0] + [lindex $bbox 2])/2]
    set visfract [expr [lindex [$data(canvas) xview] 1] - [lindex [$data(canvas) xview] 0]]
    set fract [expr double($mid - [lindex $region 0]) / ([lindex $region 2] - [lindex $region 0]) - $visfract/2.0]
    $data(canvas) xview moveto $fract

    # fr y
    set mid [expr ([lindex $bbox 1] + [lindex $bbox 3])/2]
    set visfract [expr [lindex [$data(canvas) yview] 1] - [lindex [$data(canvas) yview] 0]]
    set fract [expr double($mid - [lindex $region 1]) / ([lindex $region 3] - [lindex $region 1]) - $visfract/2.0]
    $data(canvas) yview moveto $fract
}


# PUBLIC
proc activeCanvas::selectAfterUpdate {obj group} {
    upvar 0 ::activeCanvas::$obj data
    set data(selectAfterUpdate) $group
}


# PUBLIC
proc activeCanvas::startRedraw {obj} {

    upvar 0 ::activeCanvas::$obj data
    if { [info exists data(selectAfterUpdate)] } {
        set lastSelected $data(selectAfterUpdate)
    } else {
        set lastSelected $data(selgroup)
    }
    catch {unset data}
    set data(canvas) $obj.workspace
    set data(selgroup) ""
    set data(interactionState) "objectSelection"
    set data(oldObjects) [$data(canvas) find all]
    set data(lastSelected) $lastSelected
    set data(selectedDropPointSet) ""
    set data(dropPointRotationTimer) ""
    #xxx1
}


# PUBLIC 
proc activeCanvas::getBBox {obj} {
    upvar 0 ::activeCanvas::$obj data
    eval $data(canvas) delete $data(oldObjects)
    set data(oldObjects) ""
    set bbox [$data(canvas) bbox all]
    return $bbox
}



# PUBLIC 
proc activeCanvas::endRedraw {obj} {
    # Should be called after all objects are created. The function computes the bounding box
    # for the scrollbars.

    upvar 0 ::activeCanvas::$obj data
    set bbox [getBBox $obj]

    if { ![empty $bbox] } {
        # in case canvas is not empty: compute scrollregion
        assign "x1 y1 x2 y2" $bbox
        $data(canvas) config -scrollregion "0 0 [expr $x2+10] [expr $y2+10]"
    }
    $data(canvas) raise level-1
    $data(canvas) raise level-2

    if { ![empty $data(lastSelected)] && [containsGroup $obj $data(lastSelected)] } {
        setSelection $obj $data(lastSelected)
    }
}


# PUBLIC 
proc activeCanvas::saveAsPostscript {obj filename} {
    upvar 0 ::activeCanvas::$obj data
    set bbox [$data(canvas) bbox print-1]
    $data(canvas) postscript -x [lindex $bbox 0] -y [lindex $bbox 1] \
            -width [expr [lindex $bbox 2] - [lindex $bbox 0]] \
            -height [expr [lindex $bbox 3] - [lindex $bbox 1]] \
            -file $filename 
}

