
# Modul: c.tcl
#
# Vervollstndigt die Schnittstelle zu der generierten c-Datenstruktur
# um weitere Funktionen.


# Hier die Basis-Funktionen, die in C implementiert sind.
# Siehe auch:
#  modelGen.tclinter
#  ObservedData.tclinter
#
# Fr alle Sprachelemente:
#   XXX_new
#   XXX_type
#
# Von ObservedData.fw:
#
# VLString getNameOfType(LanguageElementTypeDescriptor *x) -> METHOD;
# VLString getSubtypes(LanguageElementTypeDescriptor *x) -> METHOD;
# VLString getSupertypes(LanguageElementTypeDescriptor *x) -> METHOD;
# VLInt getIsAbstract(LanguageElementTypeDescriptor *x) -> METHOD;
# VLBoolean isSubtypeOf(LanguageElementTypeDescriptor *x, LanguageElementTypeDescriptor *other) -> METHOD;
#
# ChangeHandler *ChangeHandler_new(VLValue cmd) -> CONSTRUCTOR ChangeHandler;
#
# VLBoolean isLanguageObject(StructureObject *obj) -> METHOD;
# VLBoolean isVirtualLanguageObject(StructureObject *obj) -> METHOD;
# VLBoolean isEOLObject(StructureObject *obj) -> METHOD;
# VLBoolean isRelationObject(StructureObject *obj) -> METHOD;
#
# VLString getClassName(StructureObject *obj) -> METHOD;
# VLList getSubElements(StructureObject *obj) -> METHOD;
# StructureObject *getParent(StructureObject *obj) -> METHOD;
# void addChangeHandler(StructureObject *obj, ChangeHandler *x) -> METHOD;
# void removeChangeHandler(StructureObject *obj, ChangeHandler *x) -> METHOD;
# void propagateChange(StructureObject *obj) -> METHOD; 
# VLString getInfo(StructureObject *obj) -> METHOD;
# StructureObject *getRoot(StructureObject *obj) -> METHOD;
# VLBoolean isAncestorOf(StructureObject *a, StructureObject *b) -> METHOD;
#
# void insertBefore_internal(VirtualLanguageObject *obj, LanguageObject *elem) -> METHOD insertBefore;
# VirtualLanguageObject *getSuccessor(VirtualLanguageObject *obj) -> METHOD;
# VirtualLanguageObject *getPredecessor(VirtualLanguageObject *obj) -> METHOD;
#
# VLBoolean canExtract(LanguageObject *obj) -> METHOD;
# void extract(LanguageObject *obj) -> METHOD;
# LanguageElementTypeDescriptor *getTypeId(LanguageObject *obj) -> METHOD;
# VLList getRefererList(LanguageObject *obj) -> METHOD;
#
# VLString getRole(RelationObject *obj) -> METHOD;
# VLBoolean isAggregate(RelationObject *obj) -> METHOD;
#
# LanguageElementTypeDescriptor *getAcceptedType(AggregateRelation *obj) -> METHOD;
# EOLObject *getEOL(AggregateRelation *obj) -> METHOD;
# VirtualLanguageObject *getFirst(AggregateRelation *obj) -> METHOD;
# VLBoolean maxMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
# VLBoolean minMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
# VLBoolean minProcessorMultiplicityIsOne(AggregateRelation *obj) -> METHOD;
#
# VLValue getValue(ObservedType *obj) -> METHOD;
# void setValue_internal(ObservedType *obj, VLValue value) -> METHOD setValue;
# VLString getTypeName(ObservedType *obj) -> METHOD;

# --- class hierarchy  ----------------------------------------------------------------

# Klassenzugehrigkeit kann mit den Abfragefunktionen isXXX abgefragt werden.
#
# -StructureObject
#   -VirtualLanguageObject
#     -LanguageObject ("objectNode")
#     -EOLObject
#   -RelationObject ("attrNode")
#     -Leaf
#       -Value
#       -Reference
#     -Aggregate



proc c::isStructureObject {obj} {
    # for the sake of completeness
    return 1
}


proc c::isLeaf {obj} {
    return [equal [c::getClassName $obj] ObservedType]
}


proc c::isTerminal {obj} {
    # OBSOLETE!
    return [isLeaf $obj]
}


proc c::isValue {obj} {
    return [expr {[c::isLeaf $obj] && ![equal [c::getTypeName $obj] VLPointer]}]
}


proc c::isReference {obj} {
    return [expr {[c::isTerminal $obj] && [equal [c::getTypeName $obj] VLPointer]}]
}


proc c::isAggregate {obj} {
    return [equal [c::getClassName $obj] AggregateRelation]
}



# --- change operations  ----------------------------------------------------------------

# ACHTUNG! Diese Funktionen sollten nie direkt aufgerufen werden.
# Statt dessen sollten die Funktionen im Modul "change" benutzt werden.
#


namespace eval d {}



proc d::getFile {node} {
    variable fileArr
    return $fileArr([c::getRoot $node])
}


proc d::setFile {node file} {
    variable fileArr
    set fileArr($node) $file
}


proc d::createObject {className file} {
    set objNode [c::${className}_new]
    setFile $objNode $file
    return $objNode
}


proc d::createRoot {} {
    set objNode [c::Root_new]
    setFile $objNode $objNode
    return $objNode
}


proc d::setValue {obj value} {
    c::setValue_internal $obj $value
}


proc d::insert {agg newElem} {
    #puts "insert [c::getContext $agg] [c::getContext $newElem]"
    c::insertBefore_internal $agg $newElem
}


proc d::delete {obj} {
    file::beforeDeletion [file::getFileOfRoot [getFile $obj]] $obj
    # xxx: really delete object!
}


proc d::extract {obj} {
    c::extract $obj
}


proc d::copy {obj file} {
    # Erzeugt eine exakte Kopie von $obj.
    # Die neue Kopie gehrt zu $file.
    # Interne Referenzen der Kopie zeigen auf kopierte Knoten.
    # Externe Referenzen bleiben erhalten.

    set newobj [copy_copyStructureRec $obj oldToNew $file]
    copy_changeReferencesRec $newobj oldToNew
    return $newobj
}


proc d::copy_copyStructureRec {obj oldToNewName file} {
    # private
    upvar $oldToNewName oldToNew
    set newobj [createObject [c::getClassName $obj] $file]
    set oldToNew($obj) $newobj
    foreach sub [c::getSubElements $obj] {
        set newsub [c::getSubElement $newobj [c::getRole $sub]]
        if [c::isValue $sub] {
            setValue $newsub [c::getValue $sub]
        } elseif [c::isReference $sub] {
            setValue $newsub [c::getValue $sub]
        } elseif [c::isAggregate $sub] {
            foreach subsub [c::getSubElements $sub] {
                insert [c::icBelow $newsub] [copy_copyStructureRec $subsub oldToNew $file]
            }
        } else {
            assert 0
        }
    }
    return $newobj
}


proc d::copy_changeReferencesRec {obj oldToNewName} {
    # private
    upvar $oldToNewName oldToNew
    if { [c::isReference $obj] && [info exists oldToNew([c::getValue $obj])] } {
        setValue $obj $oldToNew([c::getValue $obj])
    }
    foreach sub [c::getSubElements $obj] {
        copy_changeReferencesRec $sub oldToNew
    }
}



# --- insertion contexts  ------------------------------------------------------------


proc c::icBelow {aggObj} {
    return [getEOL $aggObj]
}

proc c::icAfter {elem} {
    return [getSuccessor $elem]
}

proc c::icBefore {elem} {
    return $elem
}


# --- pretty printing ----------------------------------------------------------------

proc c::getContext {obj} {
    if [c::isLanguageObject $obj] {
        return [c::getClassName $obj]
    } elseif [c::isRelationObject $obj] {
        return "[c::getClassName [c::getParent $obj]].[c::getRole $obj]"
    } elseif [c::isEOLObject $obj] {
        return "EOL_OBJECT"
    } else {
        return "UNKNOWN_OBJECT"
    }
}


# Better name: getSuperclasses
proc c::getContexts {node} {
    set result ""
    if [c::isLanguageObject $node] {
        set obj $node
        set role ""
    } else {
        set obj [getParent $node]
        set role ".[c::getRole $node]"
    }
    foreach type [getSupertypes [getTypeId $obj]] {
        lappend result "[getNameOfType $type]$role"
    }
    return $result
}


proc c::getObjectRep {node} {
    if [isLanguageObject $node] {
        return "[getContext $node] '[basiccomp::getFullName $node]'"
    } elseif [c::isTerminal $node] {
        set class [c::getClassName [c::getParent $node]]
        set name [basiccomp::getFullName [c::getParent $node]]
        return "ATTR [c::getRole $node] OF $class '$name'"
    } else {
        return "XXX"
    }
}


proc c::getComment {obj} {
    if [isVirtualLanguageObject $obj] {       
        if [getSubElement $obj name -var nameObj] {
            set name [getValue $nameObj]
            if [empty $name] {
                set name "~"
            }
        } else {
            set name "\#"
        }
        return "${name}([getClassName $obj])"
    } elseif [isRelationObject $obj] {
        return "[getComment [getParent $obj]].[getRole $obj]([getClassName $obj])"
    } else {
        assert 0
    }
}


proc c::getCommentedObject_unused {obj} {
    return "$obj {[getComment $obj]}"
}


proc c::print_unused {elem {prefix ""}} {
    if [c::isLanguageObject $elem] {
        puts "$prefix [c::getClassName $elem]"
    } else {
        puts "$prefix [c::getRole $elem] ([c::getInfo $elem])"
    }
    foreach sub [c::getSubElements $elem] {
        print $sub "$prefix  "
    }
}



# --- Rest ---------------------------------------------------------------------------


proc c::startProcessor {name root} {
    vlProperty::resetForEvaluator
    vlAccess::setRootNode $root
    startProcessor_$name $root
    c::cleanupTemp
}


proc c::findReferences {targetObj root} {
    set result ""
    foreach ref [getRefererList $targetObj] {
        if [isAncestorOf $root $ref] {
            lappend result $ref
        }
    }
    return $result
}


proc c::getAttr {obj name} {
    return [getValue [getSubElement $obj $name]]
}


proc c::getAggregateSubElements {obj} {
    set result ""
    foreach sub [c::getSubElements $obj] {
        if [c::isAggregate $sub] {
            lappend result $sub
        }
    }
    return $result
}


proc c::getSubElement {obj name args} {
    assert {![nullptr $obj]}

    handleArgs [list -var ""]
    if ![empty $arg(var)] {
        upvar $arg(var) result
    }
    set result ""
    set subs [c::getSubElements $obj]
    foreach sub $subs {
        if { [string compare $name [c::getRole $sub]]==0 } {
            set result $sub
            break
        }
   }
   if [empty $arg(var)] {
       if [empty $result] {
           error "No sub element '$name' in '[c::getClassName $obj]'"
       } else {
           return $result
       }
   } else {
       return [expr ![empty $result]]
   }
}


proc c::nodeIsSubtypeOf {node type} {
    if ![isLanguageObject $node] {
        return 0
    }
    return [isSubtypeOf [getTypeId $node] [${type}_type]]
}


proc c::findIncluding {type node} {
    set orig $node
    while 1 {
        if [nullptr $node] {
            error "Severity 1: Can't find including $type of node $orig"
        }
        if [nodeIsSubtypeOf $node $type] {
            return $node
        }
        set node [c::getParent $node]
    }
}


proc c::getList {path} {
    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]

    regexp {([^.]+).(.+)} $path dummy obj tail
    # todo: effciency
    return [treePath::eval $obj "THIS.$tail"]
}


proc c::evalSimplePathExpression {obj path} {
    # Format: THIS-Style
    # Called by vlGet in vlAccess.c

    # Transforms into other format
    regexp {THIS\.(.+)} $path dummy tail

    set error [catch {
        set result [c::get $obj.$tail]
    }]
    if { $error } {
        log::error "Error while evaluating '$obj.$tail'. Look for String '***ERROR***' in generated output!"
        set result "***ERROR***"        
    }
    return $result
}   


proc c::evalSimplePathExpression2 {path} {
    # Format: Obj-Style
    # Codes:
    # - Severity 1: Null-Ref, Fehlendes Including
    # - Severity 2: Nicht existierendes Attribut, Child liefert mehr als einen Wert
    # - Severity 3: Syntaxfehler

    set list [split $path "."]
    set curr [lindex $list 0]
    foreach jump [lrange $list 1 end] {
        switch -glob $jump {
            "PARENT" {
                set curr [c::getParent $curr]
            }
            "VALUE" {
                if { [nullptr $curr] } {
                    error "Severity 1: .VALUE applied to NULL-Reference"
                } else {
                    set curr [c::getValue $curr]
                }
            }
            "CHILD" {
                if { [nullptr $curr] } {
                    error "Severity 1: .CHILD applied to NULL-Reference"
                }
                set subList [c::getSubElements $curr]
                if { [llength $subList] > 1 } {
                    error "Severity 2: .CHILD yields more than one value"
                }
                if { [llength $subList] == 1 } {
                    set curr [lindex $subList 0]
                } else {
                    set curr "0x0"
                }
            }
            "SEMANTIC" {
                if [getModel $curr model] {
                    set curr $model
                } else {
                    error "Severity 1: .SEMANTIC yields NULL"
                }                
            }
            "INCLUDING *" {
                set curr [findIncluding [lindex $jump 1] $curr]
            }
            "SEMANTICREF" {
                if [getSubElement $curr modelRef -var jump] {
                    set curr $jump
                } else {
                    error "Severity 2: .SEMANTICREF yields NULL"
                }
            }
            default {
                if [regexp {^\^(.*)$} $jump dummy type] {
                    set curr [c::findIncluding [c::getParent $curr]]
                } else {
                    set curr [c::getSubElement $curr $jump]
                }
            }
        }
    }
    return $curr
}


proc c::get {path} {
    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]

    set err [catch {evalSimplePathExpression2 $path} result]
    if { $err } {
        error "c::get: Error while evaluating '$path': $result"
    } else {
        return $result
    }
}


proc c::tryGet {path resultName} {
    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]
    upvar $resultName res

    set err [catch {evalSimplePathExpression2 $path} result]
    if { $err } {
        if { [regexp "^Severity 1:" $result] } {
            return 0
        } else {
            error "c::get: Error while evaluating '$path': $result"
        }
    } else {
        set res $result
        return 1
    }
}


proc c::getOrNull {path} {
    set path [uplevel 1 [list subst -nobackslashes -nocommands $path]]
    if [tryGet $path result] {
        return $result
    } else {
        return 0x0
    }
}    




proc c::traverseTreeRec {root classCheck varName op upLevel} {
    if [$classCheck $root] {
        upvar $upLevel $varName var
        set var $root
        uplevel $upLevel $op
        if [empty $var] {
            set root ""
        }
    }
    if ![empty $root] {
        foreach sub [c::getSubElements $root] {
            traverseTreeRec $sub $classCheck $varName $op [expr $upLevel+1]
        }
    }
}


proc c::traverseTree {root classCheck varName op} {
    traverseTreeRec $root $classCheck $varName $op 2
}



proc c::getModel {obj resultName} {
    # true  - wenn $obj ein Modellelement repraesentiert. In diesem Fall enthaelt
    #         die Variable $resultName das Modellelement
    # false - sonst

    upvar $resultName result

    if [nullptr $obj] {
        return 0
    }
    if [c::isRelationObject $obj] {
        if ![getModel [c::getParent $obj] parmodel] {
            return 0
        }
        if ![c::getSubElement $parmodel [c::getRole $obj] -var model] {
            return 0
        }
        set result $model
        return 1
    } elseif [c::isLanguageObject $obj] {
        if ![c::getSubElement $obj modelRef -var modelRef] {
            return 0
        }
        set model [c::getValue $modelRef]
        if [nullptr $model] {
            return 0
        }
        set result $model
        return 1
    } elseif [c::isEOLObject $obj] {
        if ![getModel [getParent $obj] parentModel] {
            return 0
        }
        set result [getEOL $parentModel]
        return 1
    } else {
        error "Unknown object type: [c::getClassName $obj]"
    }
}


proc c::getModelClass {repClass} {
    return [extractSubstring {[^_]+_(.*)} $repClass]
}


proc c::getRepClass {modelClass viewName} {
    return "${viewName}_${modelClass}"
}


proc c::siblings {a b} {
    return [equal [getParent $a] [getParent $b]]
}


proc c::isRoot {obj} {
    return [nullptr [c::getParent $obj]]
}



proc c::getSubElementCount {obj} {
    return [llength [getSubElements $obj]]
}


proc c::getTypeIdFromName {typeName} {
    return [${typeName}_type]
}


proc c::hasClass {class} {
    return [procExists "c::${class}_new"]
}


proc c::canInsertType {insertionContext typeId} {
    set agg [c::getParent $insertionContext]
    return [expr {
      [isSubtypeOf $typeId [getAcceptedType [getParent $insertionContext]]] &&
      !([maxMultiplicityIsOne $agg] && [getSubElementCount $agg]==1)
  }]     
}

proc c::canInsert {insertionContext newElem} {
    return [expr {
      ![isAncestorOf $newElem $insertionContext] &&
      [canInsertType $insertionContext [getTypeId $newElem]]
  }]
}


proc c::getObjCount {objNode} {
    set objCount 0
    traverseTree $objNode isLanguageObject subObjNode {
        incr objCount
    }
    return $objCount    
}


proc c::getClassOfChild {aggNode} {
    # Liefert den Klassennamen eines SUB?-Attributs oder den leeren String
    set sub [c::getList $aggNode.CHILDREN]
    switch [llength $sub] {
        0 {
            return ""
        }
        1 {
            return [getClassName [lindex $sub 0]]
        }
        default {
            error "c::getClassOfChild called for SUB*-Attribute"
        }
    }
}


proc c::getAcceptedClassesForType {typeDescriptor} {
    set result ""
    foreach subtype [getSubtypes $typeDescriptor] {
        if ![getIsAbstract $subtype] {
            eval lappend result [getNameOfType $subtype]
        }
    }
    return $result
}


proc c::getAcceptedClassesForChild {aggNode} {
    return [getAcceptedClassesForType [getAcceptedType $aggNode]]
}



# ---- Persistent IDs -----------------------------------------------------------------------------------


# PUBLIC
proc c::setPersistentId {file node id} {
    variable data
    if { ![info exists data(node2id,$file-$node)] && ![info exists data(id2node,$file-$id)] } {
        set data(node2id,$file-$node) $id
        set data(id2node,$file-$id) $node
    }
}


# PUBLIC
proc c::getPersistentId {file node} {
    variable data
    if { ![info exists data(node2id,$file-$node)] } {
        set id [getNewPersistentId $file]
        set data(node2id,$file-$node) $id
        set data(id2node,$file-$id) $node
    }
    return $data(node2id,$file-$node)
}


# PRIVATE
proc c::getNewPersistentId {file} {
    variable data
    set maxId 1000000
    while 1 {
        set candidate [expr { $maxId + int($maxId * rand()) }]
        if { ![info exists data(id2node,$file-$candidate)] } {
            return $candidate
        }
    }
}
