
namespace eval change {}

# Modul: change.tcl
#
# Dieses Modul implementiert elementare Editieroperationen und
# protokolliert sie, um sie bei Bedarf rckgngig machen zu knnen.
#
# Achtung! Editieroperationen des Benutzers sollten auf Funktionen des Moduls
# "edit" abgebildet werden.
#


# --- verwaltungsoperationen -------------------------------------------------------


# PUBLIC
proc change::getEditState {file} {
    upvar \#0 change::data-$file data
    return [llength $data(undoList)]
}


# PUBLIC
proc change::addCheckpoint {file} {
    upvar \#0 change::data-$file data
    lappend data(checkpoints) [llength $data(undoList)]
    log::log change "file $file: addCheckpoints -> $data(checkpoints)"
}


# PUBLIC
proc change::undo {file} {
    upvar \#0 change::data-$file data

    foreach restoreState [lreverse $data(checkpoints)] {
        if { $restoreState < [llength $data(undoList)] } {
            restoreState $file $restoreState
            return 1
        }
    }
    return 0
}


# PUBLIC
proc change::clipCheckpoints {file maxValue} {
    upvar \#0 change::data-$file data

    set oldCheckpoints $data(checkpoints)
    set data(checkpoints) ""
    foreach c $oldCheckpoints {
        if { $c < $maxValue } {
            lappend data(checkpoints) $c
        }
    }
    lappend data(checkpoints) $maxValue
    log::log change "file $file: clipCheckpoints -> $data(checkpoints)"
}


# PUBLIC
proc change::restoreState {file restoreState} {
    upvar \#0 change::data-$file data
    log::log change "file $file: Restoring edit state $restoreState"

    if { $restoreState > [llength $data(undoList)] } {
        error "Undo: Invalid restoreState!"
    }
    if { $restoreState == [llength $data(undoList)] } {
        return
    }

    foreach operation [lreverse [lrange $data(undoList) $restoreState end]] {
        eval $operation
    }
    set data(undoList) [lreplace $data(undoList) $restoreState end]
    clipCheckpoints $file [llength $data(undoList)]

    log::log change "New checkpoints: $data(checkpoints)"
}


# PUBLIC
proc change::getFile {node} {
    return [d::getFile $node]
}


# PUBLIC
proc change::setFile {node file} {
    return [d::setFile $node $file]
}


# --- intern ----------------------------------------------------------------------------


# PRIVATE
proc change::getObjectRep {node} {
    return "$node\[[c::getContext $node]\]"
}


# PRIVATE
proc change::addUndoOperation {file cmd} {
    upvar \#0 ::change::data-$file data
    lappend data(undoList) $cmd
}



# --- editieroperationen ------------------------------------------------------------------

# PUBLIC
proc change::initUndoList {file} {
    upvar \#0 ::change::data-$file data
    set data(undoList) ""
    set data(checkpoints) ""
}

# PUBLIC
proc change::createRoot {} {
    set file [d::createRoot]
    initUndoList $file
    log::log change "file $file: createRoot"
    return $file
}


# PUBLIC
proc change::createObject {class file} {
    set node [d::createObject $class $file]
    log::log change "file $file: createObject $class -> $node"
    addUndoOperation $file [list d::delete $node]
    return $node
}


# PUBLIC
proc change::copy {objNode destFile} {
    set node [d::copy $objNode $destFile]

    set file [getFile $objNode]
    log::log change "file $file: copy $objNode -> $node (file $destFile)"
    return $node
}


# PUBLIC
proc change::setValue {valNode value} {
    if [equal [c::getValue $valNode] $value] {
        return
    }
    set file [getFile $valNode]

    log::log change "file $file: setValue [getObjectRep $valNode] $value"
    addUndoOperation $file [list d::setValue $valNode [c::getValue $valNode]]
    d::setValue $valNode $value
}


# PUBLIC
proc change::insert {insContext newElem} {
    if ![c::isVirtualLanguageObject $insContext] {
        error "not a VirtualLanguageObject: [c::getContext $insContext]"
    }

    extract $newElem

    set file [getFile $insContext]

    log::log change "file $file: insert [getObjectRep $insContext] [getObjectRep $newElem]"
    addUndoOperation $file [list d::extract $newElem]
    d::insert $insContext $newElem
}


# PUBLIC
proc change::extract {objNode} {

    if [nullptr [c::getParent $objNode]] {
        return
    }
    set file [getFile $objNode]

    log::log change "file $file: extract [getObjectRep $objNode]"
    addUndoOperation $file [list d::insert [c::getSuccessor $objNode] $objNode]
    d::extract $objNode
}


# PUBLIC
proc change::delete {objNode} {

    set file [getFile $objNode]

    extract $objNode
    foreach ref [c::getRefererList $objNode] {
        setValue $ref "0x0"
    }

    log::log change "file $file: delete [getObjectRep $objNode]"
    d::delete $objNode
}


# --- convenient functions ----------------------------------------------------


proc change::mkTree {class file args} {
    # Shortcut function to create tree structures.
    # Used by "gded".

    set obj [createObject $class $file]
    foreach {attr value} $args {
        if ![regexp {^-(.*)$} $attr dummy attr] {
            log::error "Must begin with a dash: $attr"
        } else {
            setValue [c::get $obj.$attr] $value
        }
    }
    return $obj
}


proc change::setAttr {obj attr value} {
    set sub [c::getSubElement $obj $attr]
    setValue $sub $value
}
