
# Module: file.tcl
#
# Reprsentiert einen Strukturbaum. Untersttzt das Laden und Speichern,
# verwaltet Sichten darauf und stoesst Aktualisierungen dieser Sichten an.


namespace eval file {
    set cdata(files) ""
    set cdata(lastUnnamedNumber) 0
}


### klassenfunktionen

proc file::findNodeByInternalId {} {
    if [CDlg::getString [i18n::text "Find node by internal ID"] node] {
        highlightNodeInTreeView $node
    }
}


proc file::getInitialDir {id} {
    variable initialDir
    if [info exists initialDir($id)] {
        return $initialDir($id)
    } else {
        return "."
    }
}

proc file::setInitialDir {id location} {
    variable initialDir
    set initialDir($id) $location
}


proc file::getProcessorNames {} {
    if [info exists ::file::processorNames] {
        return $::file::processorNames
    } else {
        return ""
    }
}

proc file::registerFile {oid} {
    variable cdata
    lappend cdata(files) $oid
}

proc file::unregisterFile {oid} {
    variable cdata
    lremove cdata(files) $oid
}

proc file::getAllFiles {} {
    variable cdata
    return $cdata(files)
}


proc file::highlightNodeInTreeView {node} {
    set oid [getFileOfNode $node]
    oop::data $oid
    set view [openView $oid [getRoot $oid] "Tree view"]
    treeView::highlightNode $view $node
}


proc file::highlightNode {node} {
    set oid [getFileOfNode $node]
    oop::data $oid
    
    # find suitable view node
    set viewNode $node
    while { ![nullptr $viewNode] } {
        set list [treedraw::getViewNames $viewNode]
        if ![empty $list] {
            set viewType [lindex $list 0]
            break
        } else {
            set viewNode [c::getParent $viewNode]
        }
    }

    if { ![nullptr $viewNode] } {
        set view [openView $oid $viewNode $viewType]
        oop::call selectNode $view $node
    } else {
        highlightNodeInTreeView $node
    }
}


proc file::getFileOfRoot {root} {
    variable cdata
    return $cdata(fileOfRoot-$root)
}

proc file::getFileOfNode {node} {
    return [getFileOfRoot [c::getRoot $node]]
}

proc file::getFiletypes {} {
    set result ""
    lappend result [list "Editor Data" $::cfg::extension]
    lappend result [list "All Files" "*"]
    return $result
}

proc file::getDefaultExtension {} {
    return $::cfg::extension
}


#### Konstruktoren


proc file::new {} {
    set oid [oop::new]
    setRoot $oid [xmlio::getNewFile]
    change::initUndoList [getRoot $oid]
    registerFile $oid
    return $oid
}

proc file::openNoWindow {filename} {
    set oid [oop::new]
    setRoot $oid [xmlio::load $filename]
    change::initUndoList [getRoot $oid]
    setFilename $oid $filename
    registerFile $oid
    return $oid
}


proc file::openWithWindow {filename} {
    set oid [openNoWindow $filename]
    if ![empty $oid] {
        openMainView $oid
    }
    return $oid
}


proc file::interactiveOpen {} {
    
    set filename [tk_getOpenFile -filetypes [getFiletypes] -defaultextension [getDefaultExtension] -initialdir [getInitialDir "xmlFiles"]] 
    if {$filename == ""} return ""
    setInitialDir "xmlFiles" [file dirname $filename]
    return [openWithWindow $filename]
}

### objektfunktionen

proc file::undo {oid} {
    if ![change::undo [getRoot $oid]] {
        CDlg::Information [i18n::text "Undo is not available."] "Undo"
    }
}


proc file::openAbstractView {oid} {
    openView $oid [getRoot $oid] "Tree view"
}


proc file::openErrorView {oid} {
    openView $oid [getRoot $oid] "Check view"
}


proc file::openMainView {oid} {
    set list [treedraw::getViewNames [getRoot $oid]]
    if [empty $list] {
        openView $oid [getRoot $oid] "Tree view"
    } else {
        openView $oid [getRoot $oid] [lindex $list 0]

    }
}


proc file::openAttributeDialog {oid viewNode} {
    set className [c::getClassName $viewNode]
    if [info exists ::treedraw::specialDialog($className)] {
        set viewType $::treedraw::specialDialog($className)
        set view [specialDialogView::new $viewNode $viewType]
        updateNow $oid
    } else {
        treedraw::openStdAttributeDialog $viewNode
    }
}


proc file::openView {oid viewNode viewType args} {
    oop::data $oid
    handleArgs "-newView 0"

    if { [MDIFrame::getReuseViews] && !$arg(newView) } {
        foreach abstractView $data(views) {
            if [oop::call isReusable $abstractView $viewNode $viewType] {
                oop::call makeVisible $abstractView 
                return $abstractView
            }
        }
    }
    switch $viewType {
        "Tree view" {
            set view [treeView::new $viewNode]
        }
        "Check view" {
            set view [checkView::new $viewNode]
        }
        default {
            set view [lidoView::new $viewNode $viewType]
        }
    }
    updateNow $oid
    return $view
}


proc file::beforeDeletion {oid node} {
    oop::data $oid
    foreach abstractView $data(views) {
        oop::call beforeDeletion $abstractView $node
    }
}

proc file::saveAs {oid filename} {
    xmlio::save [getRoot $oid] $filename
    setFilename $oid $filename
    setModified $oid 0
}


proc file::interactiveSave {oid} {
    # returns: 1, if the file was saved

    set filename [getFilename $oid]
    if [getUnnamed $oid] {
        return [interactiveSaveAs $oid]
    } else {
        saveAs $oid $filename
        return 1
    }
}


proc file::interactiveSaveAs {oid} {
    # returns: 1, if the file was saved

    set filename [tk_getSaveFile -filetypes [getFiletypes] -defaultextension [getDefaultExtension] -initialdir [getInitialDir "xmlFiles"]]
    if {$filename == ""} {return 0}
    setInitialDir "xmlFiles" [file dirname $filename]
    saveAs $oid $filename
    return 1

}

proc file::getFilenameNoPath {oid} {
    return [file tail [getFilename $oid]]
}

proc file::getFilename {oid} {
    # liefert Dateiename mit komplettem Pfad.

    oop::data $oid
    return $data(filename)
}

proc file::setFilename {oid filename} {
    oop::data $oid
    set data(filename) [file join [pwd] $filename]
    set data(unnamed) 0
    updateLater $oid
}

proc file::getUnnamed {oid} {
    oop::data $oid
    return $data(unnamed)
}

proc file::setModified {oid value} {
    oop::data $oid
    set data(modified) $value
    updateLater $oid
}

proc file::getModified {oid} {
    oop::data $oid
    return $data(modified)
}


proc file::setRoot {oid root} {
    # privat!
    variable cdata
    oop::data $oid

    set cdata(fileOfRoot-$root) $oid
    set data(root) $root
    set data(updatePending) 0
    set data(syncCount) 0
    set data(views) ""
    set data(modified) 0
    set data(unnamed) 1
    incr cdata(lastUnnamedNumber)
    set data(filename) "Unnamed$cdata(lastUnnamedNumber)"
    set data(afterUpdate) ""
    set data(changeHandler) [c::ChangeHandler_new "file::onChange $oid"]
    c::addChangeHandler $root $data(changeHandler)
    #c::removeChangeHandler $data(root) $data(changeHandler)
    updateLater $oid
}

proc file::getRoot {oid} {
    oop::data $oid
    return $data(root)
}

proc file::addView {oid abstractView} {
    oop::data $oid
    lappend data(views) $abstractView
    updateLater $oid
}

proc file::getViews {oid} {
    oop::data $oid
    return $data(views)
}

proc file::removeView {oid abstractView} {
    oop::data $oid
    lremove data(views) $abstractView
}

proc file::updateIfNecessary {oid} {
    oop::data $oid
    if !$data(updatePending) {
        return
    }
    if { $data(syncCount) == 0 } {
        #prof::start
    }
        
    if $data(syncCount)>10 {
        log::error "Infinite synchronization"
        return
    }
 
    set data(updatePending) 0
    incr data(syncCount)
    log::process "SYNC $data(syncCount)" "Synchronization #$data(syncCount)"

    edit::doContextSensitiveInit
    if $data(updatePending) return

    syncutil::syncProgram $data(root)
    if $data(updatePending) return

    change::addCheckpoint [getRoot $oid]
    set data(syncCount) 0

    log::process "BASIC" "Basic computations"
    basiccomp::run $data(root)

    log::process "VIEWS" "Updating views"
    foreach abstractView $data(views) {
        log::log profile "Computing View $abstractView"
        oop::call update $abstractView
    }
    log::process "" "update ready -----------------------------------------"
    #prof::end

    eval $data(afterUpdate)
    set data(afterUpdate) ""
}


proc file::afterUpdate {oid cmd} {
    oop::data $oid
    if !$data(updatePending) {
        eval $cmd
    } else {
        set data(afterUpdate) $cmd
    }
}


proc file::updateNow {oid} {
    oop::data $oid
    set data(updatePending) 1
    updateIfNecessary $oid
}

proc file::updateLater {oid} {
    oop::data $oid
    set data(updatePending) 1
    after idle [list file::updateIfNecessary $oid]
}

proc file::onChange {oid} {
    oop::data $oid
    treePath::resetCache
    setModified $oid 1
    updateLater $oid
}

proc file::forceUpdate {oid} {
    updateLater $oid
}

proc file::startProcessor {oid processor} {
    if [equal $processor "EXPORT_HTML"] {
        exportHtml $oid [main::getParam dest]
        return 1
    } elseif [equal $processor "EXPORT_EPS"] {
        exportEps $oid [main::getParam dest]
        return 1
    } else {
        return [startLigaProcessor $oid $processor]
    }
}

proc file::startLigaProcessor {oid processor} {
    if ![elementOf $processor [getProcessorNames]] {
        log::log error "There is no processor '$processor'"
        return
    }
    set check1 [checkutil::checkProgram [getRoot $oid]]
    if ![empty $check1] {
        openErrorView $oid
        #messageView::new [getRoot $oid] "Errors" $check1
        return 0
    }
    log::process "PROCESS" "Running processor '$processor'"
    vlMessage::reset
    c::startProcessor $processor [getRoot $oid]
    if { [vlMessage::getErrorLevel] >= 0 } {
        messageView::new [getRoot $oid] "Errors" [vlMessage::getList]
        set result 0
    } else {
        set result 1
    }
    log::process "" "Processor ready -------------------------------------------"
    return $result
}


proc file::findUsages {oid targetObject} {
    set list ""
    foreach ref [c::findReferences $targetObject [getRoot $oid]] {
        lappend list [list $ref green [c::getObjectRep $ref]]
    }
    messageView::new [getRoot $oid] "Usages of object '[basiccomp::getFullName $targetObject]'" $list
}


namespace eval file {
    set search(string,value) ""
    set search(string,label) "Search pattern"
    set search(string,editor) Entry
    set search(string,check) ""

    set search(substrings,value) 1
    set search(substrings,label) "Match substrings"
    set search(substrings,editor) "Checkbox"
    set search(substrings,check) ""

    set search(regexp,value) 0
    set search(regexp,label) "Regular expression"
    set search(regexp,editor) "Checkbox"
    set search(regexp,check) ""

    set search(caseSens,value) 0
    set search(caseSens,label) "Case sensitive"
    set search(caseSens,editor) "Checkbox"
    set search(caseSens,check) ""

    set search(onlyNames,value) 0
    set search(onlyNames,label) "Only name attributes"
    set search(onlyNames,editor) "Checkbox"
    set search(onlyNames,check) ""
}


proc file::findRegexp {oid} {
    variable search
    if [CDlg::stdInputDialog "Search for value" [list string substrings regexp caseSens onlyNames] search] {

        # option processing
        if $search(caseSens,value) {
            set opt ""
        } else {
            set opt "-nocase"
        }
        if $search(regexp,value) {
            set pattern $search(string,value)
        } else {
            set pattern [quote::regexpQuote $search(string,value)]
        }
        if !$search(substrings,value) {
            set pattern "^${pattern}\$"
        }

        set list ""
        c::traverseTree [getRoot $oid] isValue obj {
            if { !$search(onlyNames,value) || [equal [c::getRole $obj] "name"] } {
                if [eval regexp $opt -- [list $pattern [c::getValue $obj]]] {
                    lappend list [list $obj green [c::getObjectRep $obj]]
                }
            }
        }
        messageView::new [getRoot $oid] "Search for '$search(string,value)'" $list
    }
}

proc ::file::close {oid} {
    foreach view [getViews $oid] {
        oop::call close $view
    }
    unregisterFile $oid
}

proc ::file::tryClose {oid} {
    if [confirmClose $oid] {
        close $oid
    }
}

proc ::file::confirmClose {oid} {
    # Returns 1, if the file should be closed, 0 otherwise.
                                                                                                            
    variable data
                                                                                                            
    if ![getModified $oid] {
        return 1
    }
                                                                                                            
    set msg ""
    append msg "[getFilename $oid]\n\n"
    append msg [i18n::text "File contains unsaved changes. Save it first?"]

    set res [CDlg::Question $msg "Close"]
    switch $res {
        yes {
            return [interactiveSave $oid]
        }
        no {
            return 1
        }
        cancel {
            return 0
        }
    }
}

proc ::file::tryCloseAll {} {
    # returns 1, if the application should be closed

    foreach file [getAllFiles] {
        if ![confirmClose $file] {
            return 0
        }
    }
    return 1
}


proc file::exportHtmlInteractive {oid} {
    set filename [tk_chooseDirectory -title "Export HTML" -initialdir [getInitialDir "exportDir"]]
    if {$filename == ""} {return 0}
    setInitialDir "exportDir" $filename
    exportHtml $oid $filename
    CDlg::Information [i18n::text "HTML export done!" "HTML export"]
}


proc file::exportEpsInteractive {oid} {
    set filename [tk_chooseDirectory -title "Export EPS" -initialdir [getInitialDir "exportDir"]]
    if {$filename == ""} {return 0}
    setInitialDir "exportDir" $filename
    exportEps $oid $filename
    CDlg::Information [i18n::text "EPS export done!" "EPS Export"]
}


proc file::exportHtml {oid filename} {
    if ![file exists $filename] {
        file mkdir $filename
    }
    exportHtml::exportHtml $oid $filename
}

proc file::exportEps {oid filename} {
    if ![file exists $filename] {
        file mkdir $filename
    }
    exportHtml::exportEps $oid $filename
}


proc file::showInfoDialog {oid} {
    lappend message "Filename: [getFilename $oid]"
    lappend message "Tree nodes: [c::getObjCount [getRoot $oid]]"
    lappend message "Modified: [lindex "no yes" [getModified $oid]]"

    CDlg::Information [join $message "\n\n"] "File Information"
}
