

# --- debug support ------------------------------------------------

proc assert {expr {msg ""}} {
    set result [uplevel 1 [list expr $expr]]
    if { $result != 1 } {
        if [empty $msg] {
            error "Assertion failed: $expr"
        } else {
            set output [uplevel 1 [list subst $msg]]
            error "Invalid state: $output"
        }
    }
}


proc ifDebug {code} {
    uplevel 1 [list eval $code]
}


proc setDebugOff {} {
    proc assert {expr {msg ""}} {}
    proc ifDebug {code} {}
}


# --- essential functions ----------------------------------------


proc max {a b} {
    if { $a > $b } {
        return $a
    } else {
        return $b
    }
}


proc empty {x} {
    return [expr { [string length $x]==0 }]
}


proc equal {a b} {
    return [expr { [string compare $a $b]==0 }]
}


proc elementOf {x set} {
    return [expr { [lsearch -exact $set $x] != -1 }]
}


# --- additional list functions -----------------------------------


proc lpush {listName elem} {
    upvar $listName list
    lappend list $elem
}


proc lpop {listName} {
    upvar $listName list
    set list [lreplace $list end end]
}


proc ltop {listName} {
    upvar $listName list
    return [lindex $list end]
}


proc lremove {listName elem} {
    upvar $listName list
    set index [lsearch -exact $list $elem]
    set list [lreplace $list $index $index]
}


proc lfilter {list expr} {
    set result ""
    foreach x $list {
        if $expr {
            lappend result $x
        }
    }
    return $result
}


proc lreverse {list} {
    set result ""
    set i [llength $list]
    while { $i > 0 } {
        incr i -1
        lappend result [lindex $list $i]
    }
    return $result
}


# --- parameter handling -----------------------------------------


proc getTagName {tag} {
    assert { "[string index $tag 0]" == "-" } {First character of option '$tag' must be a dash}
    return [string range $tag 1 end]
}


proc getParamList {arrName} {
    upvar $arrName arr
    set result ""
    foreach {param value} [array get arr] {
        lappend result -$param $value
    }
    return $result
}



proc handleParams {inputList resultArrName defaultList } {
    # Analysiert die Liste "inputList" und speichert das Ergebnis im Array "resultArrName".
    # Es sind nur die Parameter erlaubt, die in "defaultList" auftauchen.
    # Bsp fr defaultList: "-width 500 -height 400 -name Unnamed"
    # Die Striche der Optionsnamen werden abgeschnitten.
    upvar $resultArrName resultArr

    foreach {tag value} $defaultList {
        assert { [string index $tag 0] == "-" } {First character of option '$tag' must be a dash}
        set resultArr([string range $tag 1 end]) $value
    }

    foreach {tag value} $inputList {
        assert { [info exists resultArr([getTagName $tag])] } {Tag $tag is not allowed! (First character should be a dash!) }
        set resultArr([string range $tag 1 end]) $value
    }
}


proc partialHandleParams {inputList resultArrName defaultList} {
    upvar $resultArrName resultArr

    foreach {tag value} $defaultList {
        set resultArr([getTagName $tag]) $value
    }

    set remainderList ""
    foreach {tag value} $inputList {
        if ![info exists resultArr([getTagName $tag])] {
            lappend remainderList $tag $value
        } else {
            set resultArr([getTagName $tag]) $value
        }
    }
    return $remainderList
}


proc handleArgs {defaultList} {
    # Analysiert die Liste "args" und speichert das Ergebnis im Array "arg" der bergeordneten Funktion.
    # Es sind nur die Parameter erlaubt, die in "defaultList" auftauchen.
    # Bsp fr defaultList: "-width 500 -height 400 -name Unnamed"
    upvar args args
    upvar arg arg
    handleParams $args arg $defaultList 
}


proc partialHandleArgs {defaultList} {
    upvar args args
    upvar arg arg
    return [partialHandleParams $args arg $defaultList]
}


# --- name creation ----------------------------------------------

set namecounter 1000

proc newname {} {
    return [newint]
}


proc newint {} {
    incr ::namecounter
    return "X$::namecounter"
}


# --- other stuff -------------------------------------------------


proc nullptr {x} {
    return [equal $x "0x0"]
}

proc procExists {name} {
    return [expr { ![empty [info commands $name]] }]
}

proc extractSubstring {pattern input} {
    if ![regexp $pattern $input dummy result] {
        error "Can't extract substring"
    }
    return $result
}


proc range {from to} {
    set result ""
    set i $from
    while { $i < $to } {
        lappend result $i
        incr i
    }
    return $result
}


proc assign {varList valueList} {
    if { [llength $varList] != [llength $valueList] } {
        error "invalid assign"
    }
    foreach i [range 0 [llength $varList]] {
        upvar 1 [lindex $varList $i] myVar
        set myVar [lindex $valueList $i]
    }
}



proc try {cmd msg} {
    set res [catch {uplevel 1 $cmd}]
    if $res {
        error $msg
    }
}

proc objData {id} {
    set ns [namespace tail [uplevel 1 namespace current]]
    uplevel 1 [list upvar \#0 ${ns}::data-$id data]
}


proc evalPercentSubst {cmdTemplate percentFunction} {
    set cmd ""
    foreach elem $cmdTemplate {
        if [equal [string index $elem 0] "%"] {
            set subst [eval $percentFunction [list $elem]]
            if [empty $subst] {
                error "Invalid placeholder '%elem'"
            }
            lappend cmd $subst
        } else {
            lappend cmd $elem
        }
    }
    return [eval $cmd]
}


